
/****************************************************************************************
#  Copyright (c) 2011, ALES S.r.l. - Via C. Bartolomeo Piazza 13, 00161 Roma Italy      #
#  Developed with the sponsorship of the Defense Advanced Research Projects             #
#  Agency (DARPA).                                                                      #
#                                                                                       #
#  Delivered to the Government with Unlimited Rights under Contract #FA8650-10-C-7074.  #
#                                                                                       #
#  THE DATA IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR               #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,             #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE          #
#  AUTHORS, SPONSORS, DEVELOPERS, CONTRIBUTORS, OR COPYRIGHT HOLDERS BE LIABLE          #
#  FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF                   #
#  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE          #
#  DATA OR THE USE OR OTHER DEALINGS IN THE DATA.                                       #
****************************************************************************************/

/**
 * 
 */
package com.eu.ales.desyre.builder.ipprovider.meta;

import java.io.File;
import java.io.FileNotFoundException;
import java.net.URI;

import org.apache.log4j.Logger;

import com.eu.ales.desyre.builder.implprovider.ImplementationNature;
import com.eu.ales.desyre.builder.ipprovider.AbstractDesyreIPProvider;
import com.eu.ales.desyre.builder.ipprovider.DesyreIPProviderException;
import com.eu.ales.desyre.builder.wrapper.ComponentJamFileJetTemplateArgument;
import com.eu.ales.desyre.builder.wrapper.jet.ComponentJamfileTemplate;
import com.eu.ales.desyre.builder.wrapper.meta.METAJetTemplateArgument;
import com.eu.ales.desyre.builder.wrapper.meta.jet.ComponentDesyreMETAWrapperHeaderTemplate;
import com.eu.ales.desyre.builder.wrapper.meta.jet.ComponentDesyreMETAWrapperSourceTemplate;
import com.eu.ales.desyre.builder.wrapper.meta.jet.ComponentIPInterfaceTemplate;
import com.eu.ales.desyre.builder.wrapper.meta.jet.ComponentSpiritDesignConfigTemplate;
import com.eu.ales.desyre.builder.wrapper.meta.jet.ComponentSpiritDesignTemplate;
import com.eu.ales.desyre.builder.wrapper.meta.jet.ComponentSpiritInterfaceTemplate;
import com.eu.ales.desyre.configuration.DesyrePathHelper;
import com.eu.ales.desyre.model.desyre.Component;
import com.eu.ales.desyre.model.desyre.System;
import com.eu.ales.desyre.model.desyre.TimeResolutionUnit;
import com.eu.ales.desyre.model.desyre.View;
import com.eu.ales.desyre.model.desyre.util.DesyreUtils;
import com.eu.ales.desyre.utils.FileUtils;

public class DesyreMETAIPProvider extends AbstractDesyreIPProvider
{
	private static final String IMPLEMENTATION_BUNDLE_EXTENSION = ".dar";

	Logger logger = Logger.getLogger(getClass());


	/**
	 * @param tempDir
	 */
	public DesyreMETAIPProvider(DesyrePathHelper desyrePathHelper) {
		super(desyrePathHelper);
	}

	@Override
	public void synthesizeIPXact() throws DesyreIPProviderException 
	{
		METAJetTemplateArgument jta = new METAJetTemplateArgument();
		jta.component = this.component;
		jta.desyrePathHelper = this.desyrePathHelper;

		try {
			FileUtils.writeString2CodeFile(
					new ComponentSpiritInterfaceTemplate().generate(jta),
					desyrePathHelper.getComponentSpiritFile(component)
			);
		} catch (FileNotFoundException e) {
			throw new DesyreIPProviderException(e);
		}
	}

	/**
	 * @param component
	 * @param minimumTimeResolution
	 * @throws DesyreIPProviderException 
	 * @throws FileNotFoundException 
	 */
	public void setupDesyreIP(System sys, Component component, TimeResolutionUnit minimumTimeResolution) throws DesyreIPProviderException
	{
		// initialize
		this.init(sys, component, minimumTimeResolution);
		// check whether the component is hierarchical or not
		if( component.isHierarchical() )
		{
			// hierarchical component synthesis
			setupHierarchicalIP();
		}else{
			setupBehavioralIP();
		}
	}

	private void setupBehavioralIP() throws DesyreIPProviderException
	{
		// generate the ip-xact representation
		synthesizeIPXact();
		// setup implementation first
		try{
		setupImplementation();
		}catch( DesyreIPProviderException e ){
			throw new DesyreIPProviderException(e);
		}finally{
			// synthesize the wrapper
			synthesizeImplementationWrapper(ImplementationNature.CPP );
		}
		
		
	}

	protected void synthesizeImplementationWrapper( ImplementationNature  nature) throws DesyreIPProviderException
	{
		try{
			/*
			 * Generate the proper DESYRE wrapper
			 */
			for (View view : component.getView()) {

				/* generate the DESYRE wrapper */
				switch ( nature ) {
				case DLL :
					synthetizeDllWrapper(component, view, minTimeResolution);
					break;

				case CPP:
					// META: synthesize implementation wrapper
					synthesizeImplementationWrapper(system, component, view, minTimeResolution);
					break; 
				}

				synthetizeJamfile( component,  view );
			}	
		}catch(FileNotFoundException e){
			throw new DesyreIPProviderException(e);
		}
	}

	protected void setupImplementation() throws DesyreIPProviderException
	{
		/* retrieve implementation and place it in the correct place */
		String compUri = component.getUri();

		try {
			logger.debug("Retrieving implementation for component " + component.getName());
			compUri = fixURI(compUri);
			File implementation = getImplementation(URI.create(compUri));
			logger.debug("Implementation retrieved");
			File implementationDir = new File(
					desyrePathHelper.getImplementationRepository(), 
					implementation.getName().replace(IMPLEMENTATION_BUNDLE_EXTENSION, "")
			);

			if (implementation.getName().endsWith(IMPLEMENTATION_BUNDLE_EXTENSION)) 
			{				
				processImplementationFile(compUri, implementation, implementationDir);
			} else {
				implementationDir = implementation;
			}
		} catch (DesyreIPProviderException e) {
			logger.error("component URI not recognized: " + compUri + " : " + e.getMessage());
			logger.warn("Generating a stub implementation instead");

			// Generate a stub component 
			for (View view : component.getView()) {
				try {
					synthesizeImplementationStub(component, view);
				} catch (FileNotFoundException e1) {
					throw new DesyreIPProviderException(e1);
				}
			}

		} catch (RuntimeException e) {
			throw new DesyreIPProviderException(e);			
		}
	}

	

	/**
	 * @param componentFolder 
	 * @throws DesyreIPProviderException 
	 * 
	 */
	private void setupHierarchicalIP() throws DesyreIPProviderException  {

		/* prepare the call to the Jet generators */
		METAJetTemplateArgument jta = new METAJetTemplateArgument();
		jta.component = component;
		jta.desyrePathHelper = this.desyrePathHelper;
		try{
			/* generate the component spirit file */
			synthesizeIPXact();

			/* generate the design spirit file */
			FileUtils.writeString2file(
					new ComponentSpiritDesignTemplate().generate(jta), 
					desyrePathHelper.getComponentSpiritDesignFile(component)
			);

			/* generate the design configuration spirit file */
			FileUtils.writeString2file(
					new ComponentSpiritDesignConfigTemplate().generate(jta), 
					desyrePathHelper.getComponentSpiritDesignConfigurationFile(component)
			);
		}catch(Exception e){
			throw new DesyreIPProviderException(e);
		}
	}



	private void synthetizeDllWrapper(Component component, View view, TimeResolutionUnit minimumResolution) 
	throws FileNotFoundException {
		METAJetTemplateArgument jta = new METAJetTemplateArgument();
		jta.component = component;
		jta.view = view;
		jta.scaleFactor = getScaleFactor(component.getTimeResolutionUnit(),  minimumResolution);
		//		jta.scaleFactor = "1000000000000000";
		jta.desyrePathHelper = this.desyrePathHelper;

		/* Generate the IP Interface */
		FileUtils.writeString2file(
				new ComponentIPInterfaceTemplate().generate(jta),
				desyrePathHelper.getComponentIPInterfaceFile(component) //new File(componentFolder, "interface/IP_Interface.h"));
		);

		/* Generate the headers */
		//		FileUtils.writeString2file(
		//				new HSDLLViewHeaderTemplate().generate(jta),
		//				desyrePathHelper.getWrapperHeaderFile(component, view.getName()));
		//		
		//		/* Generate the API header */
		//		FileUtils.writeString2file(
		//				new HSDLLImportedApiHeaderTemplate().generate(jta),
		//				desyrePathHelper.getComponentHSImportedFile(component, view.getName())
		//		);
		//
		//		/* generate the sources file */
		//		FileUtils.writeString2file(
		//				new HSDLLViewSourceTemplate().generate(jta),
		//				desyrePathHelper.getWrapperSourceFile(component, view.getName()));

	}



	private void synthesizeImplementationWrapper(System sys, Component component, View view, TimeResolutionUnit minimumResolution) 
	throws FileNotFoundException  {

		METAJetTemplateArgument jta = new METAJetTemplateArgument();
		jta.component = component;
		jta.sys = sys;
		jta.view = view;
		jta.scaleFactor = getScaleFactor(component.getTimeResolutionUnit(),  minimumResolution);
		jta.desyrePathHelper = desyrePathHelper;

		generateMETAWrapper(component, view, jta);

	}

	/**
	 * Generate a meta wrapper
	 * @param component
	 * @param view
	 * @param jta
	 * @throws FileNotFoundException
	 */	
	private void generateMETAWrapper(Component component, View view, METAJetTemplateArgument jta) throws FileNotFoundException 
	{
		/* Generate IP Interface code */
		FileUtils.writeString2file(
				new ComponentIPInterfaceTemplate().generate(jta), 
				desyrePathHelper.getComponentIPInterfaceFile(component));

		/* Generate view header file */
		FileUtils.writeString2file(
				new ComponentDesyreMETAWrapperHeaderTemplate().generate(jta), 
				desyrePathHelper.getWrapperHeaderFile(component, view.getName()));

		/* Generate view source file */
		FileUtils.writeString2file(
				new ComponentDesyreMETAWrapperSourceTemplate().generate(jta), 
				desyrePathHelper.getWrapperSourceFile(component, view.getName()));
	}

	/*
	 * Synthesize JAM file
	 */
	private void synthetizeJamfile(Component component, View view) 
	throws FileNotFoundException {
		ComponentJamFileJetTemplateArgument jta = new ComponentJamFileJetTemplateArgument();
		jta.component = component;
		jta.view = view;
		jta.desyrePathHelper = this.desyrePathHelper;

		/* create the makefile */
		FileUtils.writeString2file(
				new ComponentJamfileTemplate().generate(jta), 
				desyrePathHelper.getComponentJamfileFile(component)
		);
	}


	private void synthesizeImplementationStub(Component component, View view) throws FileNotFoundException 
	{
		logger.debug("Stub component generation is not supported for meta: " + DesyreUtils.getQualifiedName(component));
	}

	@Override
	public boolean isProvider(Component component) 
	{
		String componentProviderUri = "NoName";
		if(component.getTarget() != null){
			if(component.getTarget().getWxURI() != null){
				componentProviderUri = component.getTarget().getWxURI().toString();
			}
			else if(component.getTarget().getBxURI() != null){
				componentProviderUri = component.getTarget().getBxURI().toString();
			}
		}
		else{
			componentProviderUri = component.getUri();
		}
		
		if(( componentProviderUri.contains("meta://") )  || (  componentProviderUri.contains("desyre://") ) )
			return true;
		else		
			return false;
		
		//Original
		//if( component.getUri().contains("meta")) 
		//	return true;
		//else		
		//	return false;
	}

}


