
/****************************************************************************************
#  Copyright (c) 2011, ALES S.r.l. - Via C. Bartolomeo Piazza 13, 00161 Roma Italy      #
#  Developed with the sponsorship of the Defense Advanced Research Projects             #
#  Agency (DARPA).                                                                      #
#                                                                                       #
#  Delivered to the Government with Unlimited Rights under Contract #FA8650-10-C-7074.  #
#                                                                                       #
#  THE DATA IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR               #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,             #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE          #
#  AUTHORS, SPONSORS, DEVELOPERS, CONTRIBUTORS, OR COPYRIGHT HOLDERS BE LIABLE          #
#  FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF                   #
#  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE          #
#  DATA OR THE USE OR OTHER DEALINGS IN THE DATA.                                       #
****************************************************************************************/

package com.eu.ales.rpyplugins.metaii2desyre.impl;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.eu.ales.desyre.model.desyre.DataType;
import com.eu.ales.desyre.model.desyre.DeclarationZone;
import com.eu.ales.desyre.model.desyre.Port;
import com.eu.ales.desyre.model.desyre.PortDirection;
import com.eu.ales.desyre.model.desyre.PortKind;
import com.eu.ales.modeltransformation.impl.DerivedMappingImpl;
import com.eu.ales.rpyplugins.MClassifier;
import com.eu.ales.rpyplugins.MModelElement;
import com.eu.ales.rpyplugins.MModelElementRef;
import com.eu.ales.rpyplugins.MStereotype;
import com.eu.ales.rpyplugins.MSysMLPort;
import com.eu.ales.rpyplugins.MTag;
import com.eu.ales.rpyplugins.utils.Logger;
import com.eu.ales.utils.AssertionUtils;
import com.eu.ales.utils.AssumptionViolation;

public class SysMLPort2Port extends DerivedMappingImpl<MSysMLPort, Port> {
	private static final Logger LOGGER = Logger.getLogger(SysMLPort2Port.class);
	public static final String IN = "In";
	public static final String OUT = "Out";
	public static final String UNSPECIFIED = "Unspecified";
	public static final String LINKS_STRUCTURAL_FEATURE = "mLinks";

	private MClassifier classifier;
	private DeclarationZone desyreDeclarationZone;

	private SysMLPort2Port(MSysMLPort self) {
		super(self);
		this.classifier = null;
		this.desyreDeclarationZone = null;
	}

	public SysMLPort2Port(MSysMLPort self, MClassifier classifier, DeclarationZone desyreDeclarationZone) {
		this(self);
		this.classifier = classifier;
		this.desyreDeclarationZone = desyreDeclarationZone;
		AssertionUtils.assumption(this.classifier != null, "Invalid null classifier argument");
		AssertionUtils.assumption(this.desyreDeclarationZone != null, "Invalid null declaration zone argument");
	}

	@Override
	public void populate() {
		DataType type = new Classifier2DataType(self.getType()).mapIfNotExists(desyreDeclarationZone);
		result.setType(type);
		result.setDirection(PortDirection.IN);
		String portDirection = self.getPortDirection();
		if (portDirection == null) {
			LOGGER.logError(classifier, "invalid null port direction in port '" + self.getName() + "'");
		} else if (portDirection.equals(IN)) {
			result.setDirection(PortDirection.IN);
		} else if (portDirection.equals(OUT)) {
			result.setDirection(PortDirection.OUT);
		} else {
			LOGGER.logError(classifier, "invalid port direction kind '" + portDirection + "' in port '" + self.getName() + "'");
		}
		for (MTag tag : self.getMAllTags()) {
			Tag2Target tag2Target = new Tag2Target(tag);
			if (tag2Target.when()) {
				result.setTarget(tag2Target.map());
			}
		}
		result.setKind(getPortKind(self));
	}

	@Override
	protected void initInheritance() {
		inherits(ModelElement2NamedElement.class);
	}

	private static final Map<String, PortKind> stereotypeToPortKindMap = new HashMap<String, PortKind>() {{
		put("DiscreteEvent", PortKind.DISCRETE_EVENT);
		put("DiscreteTime", PortKind.DISCRETE_PERIODIC);
		put("Electrical", PortKind.MODELICA_FMI);
		put("ContinuousTime", PortKind.SIMULINK);
	}};

	public static PortKind getPortKind(MSysMLPort sysMLPort) {
		AssertionUtils.assumption(sysMLPort != null, "Invalid null SysML port argument");
		PortKind portKind = getPortKind(sysMLPort.getMStereotypes(), sysMLPort);
		if (portKind == null) {
			AssertionUtils.assumption(sysMLPort.getOfClass() != null, "SysMLPort '" + sysMLPort.getName() + "' is not contained in any class");
			portKind = getPortKind(sysMLPort.getOfClass().getMStereotypes(), sysMLPort);
		}
		return portKind;
	}

	public static PortKind getPortKind(List<MStereotype> stereotypes, MModelElement context) {
		MStereotype moCC = null;
		PortKind portKind = null;
		for (MStereotype stereotype : stereotypes) {
			if (Classifier2Component.isMoCCStereotype(stereotype)) {
				if (moCC == null) {
					moCC = stereotype;
					portKind = stereotypeToPortKindMap.get(moCC.getName());
					if (portKind == null) {
						LOGGER.logError(stereotype, "Unsupported MoCC stereotype");
					}
				} else {
					LOGGER.logError(context, "Multiple MoCCs stereotypes specified");
				}
			}
		}
		if (context instanceof MSysMLPort) {
			String wxURI = Tag2Target.getWxURITagValue(((MSysMLPort)context).getOfClass());
			String authority = Tag2Target.getAuthority(wxURI);
			if (authority != null && authority.equals(Tag2Target.RHAPSODY_TARGET)) {
				portKind = PortKind.RHAPSODY_4METAII;
			}
		}
		if (portKind == null) {
			LOGGER.logError(context, "No MoCC stereotype found, returning default port kind '" + PortKind.DISCRETE_EVENT + "'");
			portKind = PortKind.DISCRETE_EVENT;
		}
		return portKind;
	}

	public static boolean isSysMLPortConnected(MSysMLPort sysMLPort) {
		AssertionUtils.assumption(sysMLPort != null, "Invalid null SysML port argument");
		for (MModelElementRef ref : sysMLPort.getMModelElementRefs()) {
			if (ref.getStructuralFeatureName().equals(LINKS_STRUCTURAL_FEATURE)) {
				return true;
			}
		}
		return false;
	}
}
