
/****************************************************************************************
#  Copyright (c) 2011, ALES S.r.l. - Via C. Bartolomeo Piazza 13, 00161 Roma Italy      #
#  Developed with the sponsorship of the Defense Advanced Research Projects             #
#  Agency (DARPA).                                                                      #
#                                                                                       #
#  Delivered to the Government with Unlimited Rights under Contract #FA8650-10-C-7074.  #
#                                                                                       #
#  THE DATA IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR               #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,             #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE          #
#  AUTHORS, SPONSORS, DEVELOPERS, CONTRIBUTORS, OR COPYRIGHT HOLDERS BE LIABLE          #
#  FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF                   #
#  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE          #
#  DATA OR THE USE OR OTHER DEALINGS IN THE DATA.                                       #
****************************************************************************************/

package com.eu.ales.rpyplugins.metaii2desyre.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.eu.ales.modeltransformation.impl.MappingImpl;
import com.eu.ales.rpyplugins.MModelElement;
import com.eu.ales.rpyplugins.MTag;
import com.eu.ales.rpyplugins.utils.Logger;
import com.eu.ales.utils.AssertionUtils;
import com.eu.ales.utils.AssumptionViolation;

public class Tag2Target extends MappingImpl<MTag, com.eu.ales.desyre.model.desyre.Target> {
	private static final Logger LOGGER = Logger.getLogger(Tag2Target.class);
	public static final String WHITEBOX_URI = "wxURI";
	public static final String BLACKBOX_URI = "bxURI";
	public static final String URI_PATH_SEPARATOR = "/";
	public static final String AUTHORITY_SEPARATOR = "://";

	public static final String DESYRE_TARGET = "desyre";
	public static final String MODELICA_TARGET = "modelica";
	public static final String MODELICA_GENERATOR_TARGET = "modelicagenerator";
	public static final String RHAPSODY_TARGET = "rhapsody";
	public static final String SIMULINK_TARGET = "simulink";

	public static final List<String> SUPPORTED_TARGET_AUTHORITY_LIST = new ArrayList<String>() {{
		add(DESYRE_TARGET);
		add(MODELICA_TARGET);
		add(RHAPSODY_TARGET);
		add(SIMULINK_TARGET);
	}};
	private static final Map<String, String> GENERATOR_TARGET_AUTHORITY_MAP = new HashMap<String, String>() {{
		put(MODELICA_GENERATOR_TARGET, MODELICA_TARGET);
	}};

	public Tag2Target(MTag self) {
		super(self);
	}

	@Override
	public boolean when() {
		return ((self.getName().equals(WHITEBOX_URI) || self.getName().equals(BLACKBOX_URI)) && self.getValue() != null && !self.getValue().isEmpty());
	}

	@Override
	public void populate() {
		AssertionUtils.assumption(when(), "Unexpected tag '" + self.getName() + "' in '" + Tag2Target.class.getSimpleName() + "' translation");
		String uri = self.getValue();
		String authority = getRawAuthority(uri);
		if (authority != null) {
			String newAuthority = GENERATOR_TARGET_AUTHORITY_MAP.get(authority);
			if (newAuthority != null) {
				result.setToGenerate(true);
				uri = uri.replaceFirst(authority, newAuthority);
				authority = newAuthority;
			}
			AssertionUtils.assumption(SUPPORTED_TARGET_AUTHORITY_LIST.contains(authority), "Cannot find authority '" + authority + "' for uri '" + self.getValue() + "' in supported authority list");
			if (self.getName().equals(WHITEBOX_URI)) {
				result.setWxURI(uri);
			} else if (self.getName().equals(BLACKBOX_URI)) {
				result.setBxURI(uri);
			}
		}
	}

	public static String[] getSplittedPath(String uri) {
		if (uri != null && uri.contains(AUTHORITY_SEPARATOR)) {
			String split[] = uri.split(AUTHORITY_SEPARATOR);
			AssertionUtils.assumption(split.length > 1, "Unexpected result of string split operation over URI '" + uri + "' w.r.t. authority separator '" + AUTHORITY_SEPARATOR + "': path expected after authority");
			return split[1].split(URI_PATH_SEPARATOR);
		} else {
			LOGGER.logError("Cannot parse authority from target URI '" + uri + "'");
		}
		return null;
	}

	public static String getAuthority(String uri) {
		String authority = null;
		if (uri != null && uri.contains(AUTHORITY_SEPARATOR)) {
			String split[] = uri.split(AUTHORITY_SEPARATOR);
			AssertionUtils.assumption(split.length > 0, "Unexpected result of string split operation over URI '" + uri + "' w.r.t. authority separator '" + AUTHORITY_SEPARATOR + "'");
			authority = GENERATOR_TARGET_AUTHORITY_MAP.get(split[0]);
			if (authority == null) {
				authority = split[0];
				if (!SUPPORTED_TARGET_AUTHORITY_LIST.contains(authority)) {
					LOGGER.logError("Unexpected authority '" + authority + "' in target URI '" + uri + "': authority is not in supported authority list");
					authority = null;
				}
			}
		} else {
			LOGGER.logError("Cannot parse authority from target URI '" + uri + "'");
		}
		return authority;
	}

	public static String getRawAuthority(String uri) {
		String authority = null;
		if (uri != null && uri.contains(AUTHORITY_SEPARATOR)) {
			String split[] = uri.split(AUTHORITY_SEPARATOR);
			AssertionUtils.assumption(split.length > 0, "Unexpected result of string split operation over URI '" + uri + "' w.r.t. authority separator '" + AUTHORITY_SEPARATOR + "'");
			authority = split[0];
			if (!GENERATOR_TARGET_AUTHORITY_MAP.keySet().contains(authority) && !SUPPORTED_TARGET_AUTHORITY_LIST.contains(authority)) {
				LOGGER.logError("Unexpected authority '" + authority + "' in target URI '" + uri + "': authority is not in supported authority list");
				authority = null;
			}
		} else {
			LOGGER.logError("Cannot parse authority from target URI '" + uri + "'");
		}
		return authority;
	}

	public static String getTagValue(MModelElement modelElement, String tagName) {
		AssertionUtils.assumption(modelElement != null, "Invalid null model element argument");
		AssertionUtils.assumption(tagName != null && !tagName.isEmpty(), "Invalid null or empty tag name argument");
		for (MTag tag : modelElement.getMAllTags()) {
			if (tag.getName().equals(tagName)) {
				if (tag.getValue() == null) {
					LOGGER.logWarning(modelElement, "tag '" + tagName + "' has null value");
				}
				return tag.getValue();
			}
		}
		return null;
	}

	public static String getBxURITagValue(MModelElement modelElement) {
		return getTagValue(modelElement, BLACKBOX_URI);
	}

	public static String getWxURITagValue(MModelElement modelElement) {
		return getTagValue(modelElement, WHITEBOX_URI);
	}

	public static boolean isModelicaTarget(MModelElement modelElement) {
		String wxURI = getWxURITagValue(modelElement);
		if (wxURI != null) {
			String targetName = getAuthority(wxURI);
			if (targetName != null && targetName.equals(MODELICA_TARGET)) {
				return true;
			}
		}
		return false;
	}

	public static boolean isSimulinkTarget(MModelElement modelElement) {
		String wxURI = getWxURITagValue(modelElement);
		if (wxURI != null) {
			String targetName = getAuthority(wxURI);
			if (targetName != null && targetName.equals(SIMULINK_TARGET)) {
				return true;
			}
		}
		return false;
	}
}