
/****************************************************************************************
#  Copyright (c) 2011, ALES S.r.l. - Via C. Bartolomeo Piazza 13, 00161 Roma Italy      #
#  Developed with the sponsorship of the Defense Advanced Research Projects             #
#  Agency (DARPA).                                                                      #
#                                                                                       #
#  Delivered to the Government with Unlimited Rights under Contract #FA8650-10-C-7074.  #
#                                                                                       #
#  THE DATA IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR               #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,             #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE          #
#  AUTHORS, SPONSORS, DEVELOPERS, CONTRIBUTORS, OR COPYRIGHT HOLDERS BE LIABLE          #
#  FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF                   #
#  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE          #
#  DATA OR THE USE OR OTHER DEALINGS IN THE DATA.                                       #
****************************************************************************************/

package com.eu.ales.rpyplugins.metaiivalidator.impl;

import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;

import com.eu.ales.rpyplugins.MClassifier;
import com.eu.ales.rpyplugins.MLink;
import com.eu.ales.rpyplugins.MModelElement;
import com.eu.ales.rpyplugins.MPackage;
import com.eu.ales.rpyplugins.MProfile;
import com.eu.ales.rpyplugins.MProject;
import com.eu.ales.rpyplugins.MStereotype;
import com.eu.ales.rpyplugins.MSysMLPort;
import com.eu.ales.rpyplugins.metaiivalidator.ValidationResults;
import com.eu.ales.rpyplugins.utils.Logger;
import com.eu.ales.rpyplugins.utils.ModelUtils;

public class MetaIIValidatorImpl {
	private static final Logger logger = Logger.getLogger(MetaIIValidatorImpl.class);
	private static final String VALIDATION_RULE_PROFILE_NAME = "ValidationRule";
	private static final String VALIDATION_RULE_PREFIX = "check";
	private static final String GENERAL_MOCC_PROFILE_NAME = "MoCC";
	private static final String PORT_DIRECTION_IN = "In";
	private static final String PORT_DIRECTION_OUT = "Out";

	private Map<Class, MStereotype> classToStereotypeMap;
	private Map<MClassifier,List<MClassifier>> generalizations;
	private MProject project;
	private ValidationResults validationResults;
	private MStereotype generalMoccStereotype;
	private MStereotype validationRuleStereotype;
	private boolean initialized;

	public MetaIIValidatorImpl(MProject project) {
		this.project = project;
		this.classToStereotypeMap = new HashMap<Class, MStereotype>();
		this.generalizations = new HashMap<MClassifier,List<MClassifier>>();
		this.validationResults = new ValidationResults();
		this.generalMoccStereotype = null;
		this.validationRuleStereotype = null;
		this.initialized = false;
	}

	private void inititialize() {
		if (!initialized) {
			if (project == null) {
				throw new Error("invalid null project");
			}
			for (MProfile profile : project.getMProfiles()) {
				if (generalMoccStereotype != null && validationRuleStereotype != null) {
					break;
				}
				getBasicStereotypes(profile);
			}
			this.initialized = true;
			if (validationRuleStereotype != null) {
				for (MProfile profile : project.getMProfiles()) {
					processStereotypes(profile);
				}
			} else {
				this.initialized = false;
				logger.logError("invalid operation, basic '" + VALIDATION_RULE_PROFILE_NAME + "' stereotype not found within project's profiles");
			}
			if (generalMoccStereotype == null) {
				this.initialized = false;
				logger.logError("invalid operation, basic '" + GENERAL_MOCC_PROFILE_NAME + "' stereotype not found within project's profiles");
			}
		}
	}

	public ValidationResults validate() {
		if (!initialized) {
			inititialize();
		}
		if (generalMoccStereotype != null) {
			validateProject(project);
		}
		return validationResults;
	}

	private void getBasicStereotypes(MModelElement modelElement) {
		for (MModelElement me : modelElement.getMNestedElements()) {
			if (me instanceof MStereotype) {
				if (((MStereotype)me).getName().equals(GENERAL_MOCC_PROFILE_NAME)) {
					generalMoccStereotype = (MStereotype)me;
				}
				if (((MStereotype)me).getName().equals(VALIDATION_RULE_PROFILE_NAME)) {
					validationRuleStereotype = (MStereotype)me;
				}
			}
			getBasicStereotypes(me);
		}
		if (generalMoccStereotype == null || validationRuleStereotype == null) {
			for (MModelElement me : modelElement.getMNestedElements()) {
				getBasicStereotypes(me);
			}
		}
	}
	
	private void processStereotypes(MModelElement modelElement) {
		if (validationRuleStereotype != null) {
			if (modelElement instanceof MStereotype) {
				MStereotype stereotype = (MStereotype)modelElement;
				ModelUtils.getAllGeneralizations(stereotype, generalizations);
				if (generalizations.get(stereotype).contains(validationRuleStereotype)) {
					List<Class> applicableClasses = ModelUtils.getAllApplicableTo(stereotype, generalizations);
					for (Class clas : applicableClasses) {
						classToStereotypeMap.put(clas, stereotype);
					}
				}
			}
			for (MModelElement me : modelElement.getMNestedElements()) {
				processStereotypes(me);
			}
		} else {
			logger.logError("invalid operation, basic '" + VALIDATION_RULE_PROFILE_NAME + "' stereotype not found within project's profiles");
		}
	}

	private void validateProject(MModelElement modelElement) {
		List<Class> allClasses = ModelUtils.getAllSuperClasses(modelElement);
		allClasses.add(modelElement.getClass());
		for (Class clas : allClasses) {
			MStereotype stereotype = classToStereotypeMap.get(clas);
			if (stereotype != null) {
				invokeCheckRule(modelElement, stereotype);
			}
		}
		for (MModelElement me : modelElement.getMNestedElements()) {
			validateProject(me);
		}
		if (modelElement instanceof MPackage) {
			for (MModelElement me : ((MPackage)modelElement).getMLinks()) {
				validateProject(me);
			}
		}
		if (modelElement instanceof MClassifier) {
			for (MModelElement me : ((MClassifier)modelElement).getMLinks()) {
				validateProject(me);
			}
		}
	}

	public void invokeCheckRule(MModelElement modelElement, MStereotype stereotype) {
		invokeCheckRule(modelElement, VALIDATION_RULE_PREFIX + stereotype.getName());
	}

	public void invokeCheckRule(MModelElement modelElement, String name) {
		try {
			Method method = this.getClass().getMethod(name,MModelElement.class);
			method.invoke(this, modelElement);
		} catch (Exception e) {
		}
	}

	// CHECK RULES
	public void checkMoCCCompatibility(MModelElement modelElement) {
		if (!initialized) {
			inititialize();
		}
		if (initialized) {
			if (modelElement instanceof MLink) {
				MLink link = (MLink)modelElement;
				MSysMLPort fromPort = link.getFromSysMLPort();
				if (fromPort == null && link.getFrom() instanceof MSysMLPort) {
					fromPort = (MSysMLPort)link.getFrom();
				} else if (fromPort == null) {
					validationResults.addValidationResult(new CheckMoCCCompatibilityResult(link, "invalid link, missing 'from' port"));
					return;
				}
				MSysMLPort toPort = link.getToSysMLPort();
				if (toPort == null && link.getTo() instanceof MSysMLPort) {
					toPort = (MSysMLPort)link.getTo();
				} else if (toPort == null) {
					validationResults.addValidationResult(new CheckMoCCCompatibilityResult(link, "invalid link, missing 'to' port"));
					return;
				}
				if (fromPort.getPortDirection().equals(PORT_DIRECTION_IN)
						&& toPort.getPortDirection().equals(PORT_DIRECTION_IN)
						&& link.getFromSysMLPort() != null && link.getToSysMLPort() != null) {
					validationResults.addValidationResult(new CheckMoCCCompatibilityResult(link, "invalid link, both ports have direction '" + PORT_DIRECTION_IN + "'"));
					return;
				} else if (fromPort.getPortDirection().equals(PORT_DIRECTION_OUT)
						&& toPort.getPortDirection().equals(PORT_DIRECTION_OUT)
						&& link.getFromSysMLPort() != null && link.getToSysMLPort() != null) {
					validationResults.addValidationResult(new CheckMoCCCompatibilityResult(link, "invalid link, both ports have direction '" + PORT_DIRECTION_OUT + "'"));
					return;
				} else if (!fromPort.getPortDirection().equals(toPort.getPortDirection())
						&& (link.getFromSysMLPort() == null || link.getToSysMLPort() == null)) {
					//validationResults.addValidationResult(new CheckMoCCCompatibilityResult(link, "invalid hierarchical link, ports must have the same direction"));
					return;
				} else if (toPort.getPortDirection().equals(PORT_DIRECTION_IN) || fromPort.getPortDirection().equals(PORT_DIRECTION_OUT)) {
					MSysMLPort tmp = fromPort;
					fromPort = toPort;
					toPort = tmp;
				}
				List<MStereotype> fromStereotypes = getMoCCStereotypes(fromPort);
				if (fromStereotypes == null) {
					fromStereotypes = getMoCCStereotypes(fromPort.getOfClass());
				}
				List<MStereotype> toStereotypes = getMoCCStereotypes(toPort);
				if (toStereotypes == null) {
					toStereotypes = getMoCCStereotypes(toPort.getOfClass());
				}
				if (fromStereotypes != null) {
					for (MStereotype fromStereotype : fromStereotypes) {
						boolean found = false;
						if (toStereotypes != null) {
							for (MStereotype toStereotype : toStereotypes) {
								if (fromStereotype == toStereotype
										|| fromPort.getPortDirection().equals(PORT_DIRECTION_IN)
										&& generalizations.get(fromStereotype).contains(toStereotype)
										|| toPort.getPortDirection().equals(PORT_DIRECTION_OUT)
										&& generalizations.get(fromStereotype).contains(toStereotype)) {
									found = true;
									break;
								}
							}
						}
						if (!found) {
							if (!fromStereotype.getName().equals("Electrical")) {
								validationResults.addValidationResult(new CheckMoCCCompatibilityResult(link, fromStereotype));
							}
						}
					}
				} else {
					validationResults.addValidationResult(new CheckMoCCCompatibilityResult(link, "missing MoCC specification on 'from' port"));
				}
			} else {
				logger.logError(modelElement, "invalid model element for check rule 'checkMoCCCompatibility'");
			}
		}
	}

	private List<MStereotype> getMoCCStereotypes(MModelElement modelElement) {
		List<MStereotype> stereotypes = modelElement.getMStereotypes();
		ListIterator<MStereotype> iter = stereotypes.listIterator();
		while (iter.hasNext()) {
			MStereotype stereotype = iter.next();
			List<MClassifier> list = generalizations.get(stereotype);
			if (stereotype != generalMoccStereotype && (list == null || !list.contains(generalMoccStereotype))) {
				iter.remove();
			}
		}
		return !stereotypes.isEmpty() ? stereotypes : null;
	}
}
