
/****************************************************************************************
#  Copyright (c) 2011, ALES S.r.l. - Via C. Bartolomeo Piazza 13, 00161 Roma Italy      #
#  Developed with the sponsorship of the Defense Advanced Research Projects             #
#  Agency (DARPA).                                                                      #
#                                                                                       #
#  Delivered to the Government with Unlimited Rights under Contract #FA8650-10-C-7074.  #
#                                                                                       #
#  THE DATA IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR               #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,             #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE          #
#  AUTHORS, SPONSORS, DEVELOPERS, CONTRIBUTORS, OR COPYRIGHT HOLDERS BE LIABLE          #
#  FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF                   #
#  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE          #
#  DATA OR THE USE OR OTHER DEALINGS IN THE DATA.                                       #
****************************************************************************************/

package com.eu.ales.rpyplugins.rpyparser.impl;

import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EStructuralFeature;

import com.eu.ales.rpyplugins.MAttribute;
import com.eu.ales.rpyplugins.MInstance;
import com.eu.ales.rpyplugins.MLiteralSpecification;
import com.eu.ales.rpyplugins.MModelElement;
import com.eu.ales.rpyplugins.MModelElementRef;
import com.eu.ales.rpyplugins.MProject;
import com.eu.ales.rpyplugins.MUnsupportedModelElement;
import com.eu.ales.rpyplugins.RpypluginsFactory;
import com.eu.ales.rpyplugins.RpypluginsPackage;
import com.eu.ales.rpyplugins.rpyparser.RpyParser;
import com.eu.ales.rpyplugins.utils.FileManager;
import com.eu.ales.rpyplugins.utils.Logger;
import com.eu.ales.rpyplugins.utils.ModelUtils;
import com.telelogic.rhapsody.core.IRPApplication;
import com.telelogic.rhapsody.core.IRPAttribute;
import com.telelogic.rhapsody.core.IRPClass;
import com.telelogic.rhapsody.core.IRPClassifier;
import com.telelogic.rhapsody.core.IRPCollection;
import com.telelogic.rhapsody.core.IRPInstance;
import com.telelogic.rhapsody.core.IRPLiteralSpecification;
import com.telelogic.rhapsody.core.IRPModelElement;
import com.telelogic.rhapsody.core.IRPProject;

/**
 * This class contains functions for reflectively parsing a Rhapsody project using Rhapsody API.
 * The parsing is based on the ECore metamodel defined in the com.eu.ales.rpyplugins.model eclipse project.
 * Each class in the ECore metamodel corresponds to a class in the Rhapsody project, with the same name except form 
 * the prefix which is changed from 'RP' to 'M', the only exception being the 'MModelElementRef' class, which is used 
 * to keep track of inverse relations between objects in the ECore model, and the 'MUnsupportedModelElement' class, 
 * which is used to store in the ECore model classes that are not yet implemented in the metamodel.
 * This relationship is taken care of by the function 'getOrCreateModelElementByReflection(...)'.
 * Each relation and attribute in the ECore metamodel corresponds to a get function of the corresponding class of the Rhapsody API.
 * This relationship is taken care of by the function 'eStructuralFeatureToGetMethod(...)'.
 * There are three exceptions to this rule, represented by the following relations: 
 * 1) 'mUnsupportedModelElement', used to keep references to 'MUnsupportedModelElement' objects;
 * 2) 'mModelElements', used to store all objects created in the ECore model;
 * 3) 'mModelElementRefs', used to store all inverse references between objects in the ECore model;
 * 4) 'mProject', used to keep a reference to the root element;
 * 5) 'uri', used to store the unique identifier of each model element object.
 * The starting function for the parsing is called 'create(...)', which instantiates an MProject an then if calls 
 * recursevely the 'addModelElementsByReflection(...)', which reflectively does the following:
 * 1) it reads all the relations and attributes (EStructuralFeature) of the current ECore object;
 * 2) for each relation and attribute it constructs and invokes the corresponding 'get' method of the Rhapsody API, 
 *    using the function 'eStructuralFeatureToGetMethod(...)';
 * 3) it reflectively creates the new ECore objects, if not already created, and sets up the appropriate relation or attribute.
 * 
 * @author leon
 *
 */
public class RpyParserImpl implements RpyParser {
	private static final Logger logger = Logger.getLogger(RpyParserImpl.class);

	private static final List<String> structuralFeaturesToSkip = Arrays.asList(
			ModelUtils.UNSUPPORTED_MODEL_ELEMENTS_RELATION,
			ModelUtils.MODEL_ELEMENT_REFS_RELATION,
			ModelUtils.MODEL_ELEMENTS_RELATION,
			ModelUtils.MODEL_ELEMENT_PREFIX.toLowerCase() + "Project", "uri");

	RpypluginsFactory rpypluginsFactory;
	RpypluginsPackage rpypluginsPackage;
	IRPApplication irpApplication;
	IRPProject irpProject;
	MProject project;

	// The modelElementMap is used to collect model elements and their URI in order to avoid creating the same element more than once
	private Map<String,MModelElement> modelElementMap;

	private void assumption(boolean condition, String msg) {
		if (!condition) {
			throw new Error(msg);
		}
	}

	public RpyParserImpl(IRPApplication application) {
		assumption(application != null, "Invalid null Rhapsody application");
		this.irpApplication = application;
		this.irpProject = this.irpApplication.activeProject();
		this.rpypluginsFactory = RpypluginsFactory.eINSTANCE;
		this.rpypluginsPackage = RpypluginsPackage.eINSTANCE;
		this.modelElementMap = new HashMap<String,MModelElement>();
		logger.setApplication(this.irpApplication);
		logger.logInfo(this.getClass().getSimpleName() + ": Successful Instantiation");
		ModelUtils.setLogger(Logger.getLogger(this.irpApplication, ModelUtils.class));
	}

	/**
	 * This function retrieves the model element from the ModelElementMap, if it exists, 
	 * otherwise it reflectively creates the corresponding model element ECore object, or an MUnsupportedModelElement object, 
	 * adds it to the project's 'mModelElements' containment relation and the ModelElementMap
	 * @param project
	 * @param irpModelElement
	 * @return
	 */
	private MModelElement getOrCreateModelElementByReflection(MProject project, IRPModelElement irpModelElement) {
		MModelElement modelElement = getModelElement(project, irpModelElement);
		if (modelElement == null) {
			String irpClassName = irpModelElement.getClass().getSimpleName();
			// remove the RP prefix from the class name and add M(eta) prefix to class name
			String className = ModelUtils.MODEL_ELEMENT_PREFIX + irpClassName.substring(2);
			modelElement = ModelUtils.createModelElementByReflection(className);
			if (modelElement == null) {
				// If model element cannot be created by reflection, then create an MUnsupportedModelElement object instead
				this.logger.logWarning(irpModelElement, "not supported");
				modelElement = this.rpypluginsFactory.createMUnsupportedModelElement();
				((MUnsupportedModelElement)modelElement).setClassName(irpModelElement.getClass().getSimpleName());
			}
			// setup generic model element attributes
			modelElement.setName(irpModelElement.getName());
			modelElement.setFullPathName(irpModelElement.getFullPathName());
			modelElement.setMProject(project);
			modelElement.setUri(getModelElementURI(modelElement.getMProject(), irpModelElement));
			// setup containment relation
			ModelUtils.setECoreRelation(project, modelElement, ModelUtils.MODEL_ELEMENTS);
			// store newly created object in the ModelElementMap
			modelElementMap.put(getModelElementURI(project, irpModelElement), modelElement);
		}
		return modelElement;
	}


	/**
	 * This function computes a unique identifier for every model element.
	 * @param project
	 * @param irpModelElement
	 * @return
	 */
	private String getModelElementURI(MProject project, IRPModelElement irpModelElement) {
		if (irpModelElement instanceof IRPProject) {
			return project.getName();
		} else if (irpModelElement instanceof IRPLiteralSpecification) {
			return project.getName() + "::" + irpModelElement.getFullPathName() + "::" + ((IRPLiteralSpecification)irpModelElement).getValue();
		} else if (irpModelElement instanceof IRPClass) {
			return project.getName() + "::" + irpModelElement.getFullPathName() + "::" + irpModelElement.getName();
		} else {
			return project.getName() + "::" + irpModelElement.getFullPathName();
		}
	}

	/**
	 * This function returns a model element, if already created, taken from the ModelElementMap based on the model element's URI.
	 * It returns null if the model element has not been created yet.
	 * @param project
	 * @param irpModelElement
	 * @return
	 */
	private MModelElement getModelElement(MProject project, IRPModelElement irpModelElement) {
		return modelElementMap.get(getModelElementURI(project, irpModelElement));
	}

	/**
	 * Returns true if modelElement is already present in the modelElementMap
	 * @param project
	 * @param irpModelElement
	 * @return
	 */
	private boolean isAlreadyDefined(MProject project, IRPModelElement irpModelElement) {
		MModelElement me = getModelElement(project, irpModelElement);
		return me != null;
	}

	/**
	 * This function returns the 'get' method of the Rhapsody API corresponding to the eStructuralFeature in the ECore metamodel  
	 * @param modelElement
	 * @param eStructuralFeature
	 * @param irpModelElement
	 * @return
	 */
	private Method eStructuralFeatureToParseMethod(MModelElement modelElement, EStructuralFeature eStructuralFeature, IRPModelElement irpModelElement) {
		String methodName = null;
		String eStructuralFeatureName = eStructuralFeature.getName();
		if (eStructuralFeatureName.startsWith(ModelUtils.MODEL_ELEMENT_PREFIX.toLowerCase()) 
				&& eStructuralFeatureName.length() > 1 && Character.isUpperCase(eStructuralFeatureName.charAt(1))) {
			methodName = ModelUtils.PARSE + eStructuralFeatureName.substring(1);
		} else if (Character.isLowerCase(eStructuralFeatureName.charAt(0))) {
			methodName = ModelUtils.PARSE + eStructuralFeatureName.substring(0,1).toUpperCase() + eStructuralFeatureName.substring(1);
		} else {
			this.logger.logTrace(modelElement, "unexpected relation '" + eStructuralFeatureName + "'");
		}
		if (methodName != null) {
			try {
				return this.getClass().getMethod(methodName, MModelElement.class, EStructuralFeature.class, IRPModelElement.class);
			} catch (NoSuchMethodException e) {
				this.logger.logTrace(irpModelElement, "unexpected exception '" + e.getClass().getSimpleName() + "' while searching for method '" + methodName + "': " + e.getMessage());
			}
		}
		return null;
	}

	/**
	 * This function returns the 'get' method of the Rhapsody API corresponding to the eStructuralFeature in the ECore metamodel  
	 * @param modelElement
	 * @param eStructuralFeature
	 * @param irpModelElement
	 * @return
	 */
	private Method eStructuralFeatureToGetMethod(MModelElement modelElement, EStructuralFeature eStructuralFeature, IRPModelElement irpModelElement) {
		String methodName = null;
		String eStructuralFeatureName = eStructuralFeature.getName();
		if (eStructuralFeatureName.startsWith(ModelUtils.MODEL_ELEMENT_PREFIX.toLowerCase()) 
				&& eStructuralFeatureName.length() > 1 && Character.isUpperCase(eStructuralFeatureName.charAt(1))) {
			methodName = ModelUtils.GET + eStructuralFeatureName.substring(1);
		} else if (Character.isLowerCase(eStructuralFeatureName.charAt(0))) {
			if (eStructuralFeatureName.equals("returnType")) {
				methodName = ModelUtils.GET + "Returns";
			} else {
				methodName = ModelUtils.GET + eStructuralFeatureName.substring(0,1).toUpperCase() + eStructuralFeatureName.substring(1);
			}
		} else {
			this.logger.logTrace(modelElement, "unexpected relation '" + eStructuralFeatureName + "'");
		}
		if (methodName != null) {
			try {
				return irpModelElement.getClass().getMethod(methodName);
			} catch (NoSuchMethodException e) {
				this.logger.logTrace(irpModelElement, "unexpected exception '" + e.getClass().getSimpleName() + "' while searching for method '" + methodName + "': " + e.getMessage());
			}
		}
		return null;
	}

	/**
	 * This function reflectively creates an ECore model element corresponding to the irpModelElement
	 * and adds it to the eStructuralFeature of modelElement
	 * @param modelElement
	 * @param irpModelElement
	 * @param eStructuralFeature
	 */
	private void addModelElementByReflection(MModelElement modelElement, IRPModelElement irpModelElement, EStructuralFeature eStructuralFeature) {
		boolean alreadyDefined = isAlreadyDefined(modelElement.getMProject(), irpModelElement);
		MModelElement modelElement2 = getOrCreateModelElementByReflection(modelElement.getMProject(), irpModelElement);
		if (modelElement2 instanceof MUnsupportedModelElement) {
			((MUnsupportedModelElement)modelElement2).setStructuralFeatureName(eStructuralFeature.getName());
			eStructuralFeature = ModelUtils.getContainmentOrReferenceRelation(modelElement, ModelUtils.UNSUPPORTED_MODEL_ELEMENTS_RELATION);
		}
		// setup reference relations
		ModelUtils.setECoreRelation(modelElement, modelElement2, eStructuralFeature);
		if (!alreadyDefined) {
			addModelElementsByReflection(modelElement2, irpModelElement);
		}
	}

	/**
	 * This function recursively and reflectively does the following:
	 * 1) it reads all the relations and attributes (EStructuralFeature) of the 'modelElement' ECore object;
	 * 2) for each relation and attribute it constructs, using the function 'eStructuralFeatureToGetMethod(...)',
	 *    and invokes the corresponding 'get' method of the Rhapsody API, 
	 * 3) it reflectively creates the new ECore objects, if not already created, and sets up the appropriate relation or attribute.
	 * 
	 * @param modelElement
	 * @param irpModelElement
	 */
	private void addModelElementsByReflection(MModelElement modelElement, IRPModelElement irpModelElement) {
		EList<EStructuralFeature> eStructuralFeatureList = modelElement.eClass().getEAllStructuralFeatures();
		for (EStructuralFeature eStructuralFeature : eStructuralFeatureList) {
			if (!structuralFeaturesToSkip.contains(eStructuralFeature.getName())) {
				// for each structural feature (relation or attribute) of modelElement, we call the corresponding 
				// 'get' method of the irpModelElement. We skip the structural features that cannot correspond to 
				// any 'get' method of the irpModelElement that are listed in the 'structuralFeaturesToSkip' list.
				Method getMethod = eStructuralFeatureToGetMethod(modelElement, eStructuralFeature, irpModelElement);
				if (getMethod != null) {
					Object returnedObj = null;
					try {
						// invoke corresponding 'get' method on the irpModelElement
						returnedObj = getMethod.invoke(irpModelElement);
						if (returnedObj instanceof IRPCollection) {
							if (eStructuralFeature.isMany()) {
								// if the return object is a collection, iterate over the collection
								IRPCollection collection = (IRPCollection)returnedObj;
								for (Object obj : collection.toList()) {
									if (obj instanceof IRPModelElement) {
										// create the ECore corresponding object and add it to the structural feature of the modelElement
										addModelElementByReflection(modelElement, (IRPModelElement)obj, eStructuralFeature);
									} else if (obj instanceof Character || obj instanceof String || obj instanceof Boolean
											|| obj instanceof Short || obj instanceof Integer || obj instanceof Long
											|| obj instanceof Float || obj instanceof Double) {
										// in case of an attribute add it directly to the structural feature
										ModelUtils.setECoreAttribute(modelElement, returnedObj, eStructuralFeature);
									} else {
										this.logger.logTrace(obj, "unexpected object");
									}
								}
							} else {
								this.logger.logTrace(irpModelElement, "unexpected collection as return type while invoking method '" + getMethod.getName() + "' corresponding to ECore strucural feature '" + eStructuralFeature.getName() + "'");
							}
						} else if (returnedObj instanceof IRPModelElement) {
							// create the ECore corresponding object and add it to the structural feature of the modelElement
							addModelElementByReflection(modelElement, (IRPModelElement)returnedObj, eStructuralFeature);
						} else if (returnedObj instanceof Character || returnedObj instanceof String || returnedObj instanceof Boolean
								|| returnedObj instanceof Short || returnedObj instanceof Integer || returnedObj instanceof Long
								|| returnedObj instanceof Float || returnedObj instanceof Double) {
							// in case of an attribute add it directly to the structural feature
							ModelUtils.setECoreAttribute(modelElement, returnedObj, eStructuralFeature);
						} else if (returnedObj != null) {
							this.logger.logTrace(irpModelElement, "unexpected object of type '" + returnedObj.getClass().getSimpleName() + "' returned while invoking method '" + getMethod.getName() + "'");
						} else {
							//this.logger.logTrace(irpModelElement, "unexpected null returned while invoking method '" + getMethod.getName() + "'");
						}
					} catch (Exception e) {
						this.logger.logTrace(irpModelElement, "unexpected exception '" + e.getClass().getSimpleName() + "' while invoking method '" + getMethod.getName() + "' with returned object of type '" + (returnedObj != null ? returnedObj.getClass().getSimpleName() : "null") + "': " + e.getMessage());
					}
				} else {
					// TODO: implement ECore MetaClass and ECore Relation and java utility that reflexively calls Rhapsody API requiring arguments
					// no time to do it now, implementing shortcut here for IRPInstance APIs with arguments (see function 'parseAttributeValue' below)
					Method parseMethod = eStructuralFeatureToParseMethod(modelElement, eStructuralFeature, irpModelElement);
					if (parseMethod != null) {
						try {
							parseMethod.invoke(this, new Object[]{modelElement, eStructuralFeature, irpModelElement});
						} catch (Exception e) {
							this.logger.logTrace(irpModelElement, "invokation of parse method '" + parseMethod.getName() + "' failed");
						}
					} else {
						this.logger.logTrace(irpModelElement, "cannot find get method for ECore structural feature '" + eStructuralFeature.getName() + "' of model element '" + modelElement.getName() + "' of type '" + modelElement.eClass().getName() + "'");
					}
				}
			}
		}
	}

	/* (non-Javadoc)
	 * The root function to create the ECore model.
	 * @see com.eu.ales.rpyplugins.modelconstruction.ModelConstruction#create()
	 */
	@Override
	public void create() {
		assumption(this.irpProject != null, "Cannot parse null Rhapsody project");
		this.project = this.rpypluginsFactory.createMProject();
		this.project.setName(this.irpProject.getName());
		this.project.setFullPathName("");
		this.project.setMProject(this.project);
		this.project.setUri(getModelElementURI(this.project, this.irpProject));
		// do hierarchical creation of model by reflection
		addModelElementsByReflection(this.project, this.irpProject);
		this.logger.logInfo(this.getClass().getSimpleName() + ": Successful Completion");
	}

	@Override
	public MProject getModel() {
		assumption(this.project != null, "Invalid null project, the parsed model has not been build: call 'create()' first");
		return this.project;
	}

	/* (non-Javadoc)
	 * Save model to xmi file
	 * @see com.eu.ales.rpyplugins.rpyparser.RpyParser#saveToFile(java.io.File)
	 */
	public void saveToFile() {
		FileManager.saveMetaIIProjectToFile(this.project, RpypluginsPackage.eNS_URI, RpypluginsPackage.eINSTANCE, logger);
	}

	public void printOut() {
		assumption(this.project != null, "Invalid null model, build model first by calling function 'create(...)'");
		logger.printOut(this.project);
	}

	// SPECIFIC PARSE METHODS
	// TODO: implement ECore MetaClass and ECore Relation and java utility that reflexively calls Rhapsody API requiring arguments
	// no time to do it now, implementing shortcut here for IRPInstance API "String getAttributeValue(String)"

	public void parseAttributeValues(MModelElement modelElement, EStructuralFeature eStructuralFeature, IRPModelElement irpModelElement) {
		logger.logTrace(irpModelElement, "executing specific parse method 'parseAttributeValues('" + modelElement.getName() + "','" + irpModelElement.getName() + "')' (quick workaround still to be fixed)");
		assumption(modelElement != null, "Invalid null target model element");
		assumption(irpModelElement != null, "Invalid null source model element");
		assumption(modelElement instanceof MInstance, "Unexpected model element type '" + modelElement.eClass().getName() + "' for argument model element '" + modelElement.getName() + "'");
		assumption(irpModelElement instanceof IRPInstance, "Unexpected model element type '" + irpModelElement.getClass().getSimpleName() + "' for argument model element '" + irpModelElement.getName() + "'");
		IRPClassifier classifer = ((IRPInstance)irpModelElement).getOtherClass();
		if (classifer != null) {
			List<Object> attributes = classifer.getAttributes().toList();
			for (Object attribute : attributes) {
				if (attribute instanceof IRPAttribute) {
					String attributeName = ((IRPAttribute)attribute).getName();
					String attributeValue = ((IRPInstance)irpModelElement).getAttributeValue(attributeName);
					if (attributeValue == null) {
						List<Object> valueSpecs = ((IRPAttribute)attribute).getValueSpecifications().toList();
						if (valueSpecs.size() > 0) {
							// TODO: fix it to support multiple value specifications
							if (valueSpecs.size() != 1) {
								logger.logError(irpModelElement, "attribute '" + attributeName + "' has multiple value specifications: multiple value attributes are not supported, taking first value only");
							}
							attributeValue = ((IRPLiteralSpecification)valueSpecs.get(0)).getValue();
						} else {
							logger.logError(irpModelElement, "attribute '" + attributeName + "' has no value specification");
						}
					}
					MAttribute attr = rpypluginsFactory.createMAttribute();
					modelElement.getMProject().getMModelElements().add(attr);
					((MInstance)modelElement).getMAttributeValues().add(attr);

					// adding inverse reference
					MModelElementRef ref = rpypluginsFactory.createMModelElementRef();
					ref.setStructuralFeatureName(eStructuralFeature.getName());
					ref.setReference(modelElement);
					attr.getMModelElementRefs().add(ref);

					attr.setName(((IRPAttribute)attribute).getName());
					attr.setMultiplicity(((IRPAttribute)attribute).getMultiplicity());
					MLiteralSpecification literalSpec = rpypluginsFactory.createMLiteralSpecification();
					modelElement.getMProject().getMModelElements().add(literalSpec);
					attr.getMValueSpecifications().add(literalSpec);
					literalSpec.setValue(attributeValue);

					// adding inverse reference
					ref = rpypluginsFactory.createMModelElementRef();
					ref.setStructuralFeatureName("mValueSpecifications");
					ref.setReference(attr);
					literalSpec.getMModelElementRefs().add(ref);
				}
			}
		}
	}
}
