
/****************************************************************************************
#  Copyright (c) 2011, ALES S.r.l. - Via C. Bartolomeo Piazza 13, 00161 Roma Italy      #
#  Developed with the sponsorship of the Defense Advanced Research Projects             #
#  Agency (DARPA).                                                                      #
#                                                                                       #
#  Delivered to the Government with Unlimited Rights under Contract #FA8650-10-C-7074.  #
#                                                                                       #
#  THE DATA IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR               #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,             #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE          #
#  AUTHORS, SPONSORS, DEVELOPERS, CONTRIBUTORS, OR COPYRIGHT HOLDERS BE LIABLE          #
#  FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF                   #
#  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE          #
#  DATA OR THE USE OR OTHER DEALINGS IN THE DATA.                                       #
****************************************************************************************/

package com.eu.ales.rpyplugins.utils;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.apache.commons.io.FileUtils;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.resource.Resource;


public class FileManager {
	private static final Logger LOGGER = Logger.getLogger(FileManager.class);
	public static final String DESYRE_RESOURCES_FOLDER = "DesyreResources";
	public static final String BATCH_FILE_EXTENSION = ".bat";
	public final static String BLACK_BOX_FOLDER = "BlackBoxComponents";
	public final static String METAII_RESOURCE_FILE_NAME = "METAII_Model";
	public final static String DESYRE_RESOURCE_FILE_NAME = "METAII_DESYRE_Model";
	public final static String MODELICA_RESOURCE_FILE_NAME = "METAII_MODELICA_Model";
	public final static String RESOURCE_FILE_EXTENSION = ".xmi";
	public final static String MODELICA_FILE_EXTENSION = ".mo";
	public static final List<String> DESYRE_LIBRARY_NAMES = new LinkedList<String>() {
		private static final long serialVersionUID = 3674518672954590206L; {
		// ThickConnectors DESYRE library for translation
		add(DESYRE_RESOURCES_FOLDER + File.separator + "DESYRE.METAII.Library");
		}};

//		private static final String METAII_DESYRE_SIMULATOR = "Simulator" + File.separator + "Simulator";
	private static final String METAII_DESYRE_SIMULATOR = "";
	private static final String DESYRE_WORKSPACE_ENVIRONMENT_VARIABLE = "DESYREII_WS";
	private static String DESYRE_WORKSPACE_PATH = null;

	public static void readDesyreIIWorkspaceEnvironmentVariable() {
		if (DESYRE_WORKSPACE_PATH == null || DESYRE_WORKSPACE_PATH.isEmpty()) {
			Map<String, String> env = System.getenv();
			DESYRE_WORKSPACE_PATH = env.get(DESYRE_WORKSPACE_ENVIRONMENT_VARIABLE);
			if (DESYRE_WORKSPACE_PATH == null | DESYRE_WORKSPACE_PATH.isEmpty()) {
				throw new Error("The environment variable '" + DESYRE_WORKSPACE_ENVIRONMENT_VARIABLE + "' has not been set");
			}
		}
	}

	public static File getBatchFile(String relativePath) {
		if (!relativePath.endsWith(BATCH_FILE_EXTENSION)) {
			relativePath += BATCH_FILE_EXTENSION;
		}
		return createDesyreWorkspaceFile(relativePath);
	}

	public static File getModelicaPackageFile(String relativePath) {
		if (!relativePath.endsWith(MODELICA_FILE_EXTENSION)) {
			relativePath += MODELICA_FILE_EXTENSION;
		}
		return createDesyreWorkspaceFile(relativePath);
	}

	public static URI createMetaIIProjectURI(String name) {
		return createDesyreWorkspaceURI(name);
	}

	public static URI createDesyreModelURI(String name) {
		return createDesyreWorkspaceURI(name);
	}

	public static URI createModelicaModelURI(String name) {
		return createDesyreWorkspaceURI(name);
	}

	public static URI createDesyreWorkspaceURI(String name) {
		File xmiFile = createDesyreWorkspaceFile(name + FileManager.RESOURCE_FILE_EXTENSION);
		URI uri = URI.createFileURI(xmiFile.getAbsolutePath());
		return uri;
	}

	public static File createDesyreWorkspaceFile(String name) {
		if (name == null && name.isEmpty()) {
			throw new Error("Invalid null or empty file name");
		}
		File xmiFile = new File(getDesyreWorkspacePath(), name);
		return xmiFile;
	}

	public static String getDesyreWorkspacePath() {
		if (DESYRE_WORKSPACE_PATH == null) {
			readDesyreIIWorkspaceEnvironmentVariable();
		}
		return DESYRE_WORKSPACE_PATH;
	}

	public static File getCurrentWorkingDir() {
		String curDir = System.getProperty("user.dir");
		File workspaceDir = new File(curDir);
		return workspaceDir;
	}

	public static File getDesyreSimulatorDir() {
		return new File(getDesyreWorkspacePath(), METAII_DESYRE_SIMULATOR);
	}

	public static File getDesyreImplementationRepository() {
		return new File(getDesyreWorkspacePath() + File.separator + BLACK_BOX_FOLDER);
	}

	public static void saveMetaIIProjectToFile(EObject obj, String pkgURI, EPackage pkgInstance, Logger logger) {
		URI uri = FileManager.createMetaIIProjectURI(FileManager.METAII_RESOURCE_FILE_NAME);
		File projectFile = new File(uri.toFileString());
		try {
			logger.logTrace("Saving model to file '" + uri.toFileString() + "'");
			Serialization.saveToFile(projectFile, obj, pkgURI, pkgInstance);
		} catch (IOException e) {
			logger.logError("IOException: cannot save model to file '" + projectFile.getAbsolutePath() + "'");
		}
	}

	public static void saveDesyreModelToFile(EObject desyreSystem, List<?extends EObject> desyreLibraries, String pkgURI, EPackage pkgInstance, Logger logger) {
		try {
			for (EObject library : desyreLibraries) {
				try {
					logger.logTrace("Saving model to file '" + library.eResource().getURI().toFileString() + "'");
					library.eResource().save(null);
				} catch (IOException e) {
					logger.logError("IOException: cannot save model to file '" + library.eResource().getURI().toFileString() + "'");
					e.printStackTrace();
				}
			}
			logger.logTrace("Saving model to file '" + desyreSystem.eResource().getURI().toFileString() + "'");
			desyreSystem.eResource().save(null);
		} catch (IOException e) {
			logger.logError("IOException: cannot save model to file '" + desyreSystem.eResource().getURI().toFileString() + "'");
			e.printStackTrace();
		}
	}

	public static void saveModelicaModelToFile(Resource resource, String pkgURI, EPackage pkgInstance, Logger logger) {
		try {
			logger.logTrace("Saving model to file '" + resource.getURI().toFileString() + "'");
			resource.save(null);
		} catch (IOException e) {
			logger.logError("IOException: cannot save model to file '" + resource.getURI().toFileString() + "'");
			e.printStackTrace();
		}
	}

	public static void concatenateFiles(List<File> filesToConcatenate, File outputFile, boolean skipMissingFiles) {
		//File finalFile = new File(outputFile.getAbsolutePath() + ".tmp");
		List<String> lines = new ArrayList<String>();
		for (File file : filesToConcatenate) {
			if (file.exists()) {
				try {
					lines.addAll(FileUtils.readLines(file));
				} catch (IOException e) {
					LOGGER.logError(e.getClass().getSimpleName() + " occurred while reading from file '" + file.getAbsolutePath() + "'");
				}
			}
		}
		try {
			FileUtils.writeLines(outputFile, lines);
		} catch (IOException e) {
			LOGGER.logError(e.getClass().getSimpleName() + " occurred while writing to file '" + outputFile.getAbsolutePath() + "'");
		}
	}
}
