
/****************************************************************************************
#  Copyright (c) 2011, ALES S.r.l. - Via C. Bartolomeo Piazza 13, 00161 Roma Italy      #
#  Developed with the sponsorship of the Defense Advanced Research Projects             #
#  Agency (DARPA).                                                                      #
#                                                                                       #
#  Delivered to the Government with Unlimited Rights under Contract #FA8650-10-C-7074.  #
#                                                                                       #
#  THE DATA IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR               #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,             #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE          #
#  AUTHORS, SPONSORS, DEVELOPERS, CONTRIBUTORS, OR COPYRIGHT HOLDERS BE LIABLE          #
#  FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF                   #
#  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE          #
#  DATA OR THE USE OR OTHER DEALINGS IN THE DATA.                                       #
****************************************************************************************/

package com.eu.ales.rpyplugins.utils;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;

import com.eu.ales.rpyplugins.MModelElement;
import com.eu.ales.rpyplugins.MModelElementRef;
import com.telelogic.rhapsody.core.IRPApplication;
import com.telelogic.rhapsody.core.IRPModelElement;

public class Logger {
	public enum Level {NONE, INFO, DEBUG, TRACE};
	private enum LogKindEnum {INFO, WARNING, ERROR, TRACE};
	private static Level level = Level.INFO;
	private static final String LOG = "Log";
	private static Map<String, Logger> stringToLoggerMap = new HashMap<String, Logger>();
	private static Map<Class, Logger> classToLoggerMap = new HashMap<Class, Logger>();

	private static IRPApplication currentIrpApplication = null;
	private IRPApplication irpApplication;
	private String loggerIdentifier;

	// The visited list is used by the printOut function to avoid looping through the ECore model data structure 
	// or printing the same object more than once 
	private List<EObject> visited;

	private Logger() {
		this.loggerIdentifier = null;
		this.irpApplication = null;
		this.visited = new ArrayList<EObject>();
	}

	private Logger(IRPApplication irpApplication) {
		this();
		this.irpApplication = irpApplication;
		if (irpApplication != null) {
			Logger.currentIrpApplication = irpApplication;
		}
	}

	public static void setLevel(Level level) {
		//Logger.level = level;
	}

	public static Logger getLogger(String name) {
		if (name == null) {
			throw new Error("Invalid null name");
		}
		Logger logger = stringToLoggerMap.get(name);
		if (logger == null) {
			logger = new Logger(Logger.currentIrpApplication);
			stringToLoggerMap.put(name, logger);
			logger.loggerIdentifier = name;
		}
		return logger;
	}

	public static Logger getLogger(IRPApplication irpApplication, String name) {
		Logger logger = getLogger(name);
		logger.irpApplication = irpApplication;
		return logger;
	}

	public static Logger getLogger(Class clas) {
		if (clas == null) {
			throw new Error("Invalid null class");
		}
		Logger logger = classToLoggerMap.get(clas);
		if (logger == null) {
			logger = new Logger(Logger.currentIrpApplication);
			classToLoggerMap.put(clas, logger);
			logger.loggerIdentifier = clas.getSimpleName();
		}
		return logger;
	}

	public static Logger getLogger(IRPApplication irpApplication, Class clas) {
		Logger logger = getLogger(clas);
		logger.irpApplication = irpApplication;
		return logger;
	}

	public void setApplication(IRPApplication irpApplication) {
		this.irpApplication = irpApplication;
		if (irpApplication != null) {
			Logger.currentIrpApplication = irpApplication;
		}
	}

	/**
	 * This function is used to print out obj's name and type before the message
	 * @param obj The object whose name and type are printed out before the message
	 * @param msg The message to print out
	 * @param kind Can be WARNING or ERROR
	 */
	private void log(Object obj, String msg, LogKindEnum kind) {
		switch (kind) {
		case INFO:
			if (level == Level.NONE) {
				return;
			}
			break;
		case WARNING:
			if (level == Level.NONE || level == Level.INFO) {
				return;
			}
			break;
		case ERROR:
			if (level == Level.NONE || level == Level.INFO) {
				return;
			}
			break;
		case TRACE:
			if (level != Level.TRACE) {
				return;
			}
			break;
		default:
			logError("Unexpected log kind '" + kind + "'");
			break;
		}
		if (msg == null) {
			msg = "";
		}
		if (obj != null) {
			if (obj instanceof IRPModelElement) {
				IRPModelElement me = (IRPModelElement)obj;
				msg = "IRP Model Element '" + me.getName() + "' of type '" + me.getClass().getSimpleName() + "': " + msg;
			} else if (obj instanceof MModelElement) {
				MModelElement me = (MModelElement)obj;
				msg = "IRP Model Element '" + me.getName() + "' of type '" + me.eClass().getName() + "': " + msg;
			} else {
				msg = "Object of type '" + obj.getClass().getSimpleName() + "': " + msg;
			}
		}
		if (loggerIdentifier != null) {
			msg = "[" + loggerIdentifier + "] " + msg;
		}
		switch (kind) {
		case INFO: msg = "*  " + kind + "   * " + msg;
			break;
		case WARNING: msg = "* " + kind + " * " + msg;
			break;
		case ERROR: msg = "*  " + kind + "  * " + msg;
			break;
		case TRACE: msg = "*  " + kind + "  * " + msg;
			break;
		}
		// PRINT OUT LOG MESSAGE
		if (this.irpApplication != null) {
			// print out log message to Rhapsody Log pane
			this.irpApplication.writeToOutputWindow(LOG , "- " + msg + " -\n");
		} else {
			// print out log message to standard output stream
			System.out.println("- " + msg + " -");
		}
	}

	public void logInfo(String msg) {
		log(null, msg, LogKindEnum.INFO);
	}

	public void logWarning(String msg) {
		log(null, msg, LogKindEnum.WARNING);
	}

	public void logError(String msg) {
		log(null, msg, LogKindEnum.ERROR);
	}

	public void logTrace(String msg) {
		log(null, msg, LogKindEnum.TRACE);
	}

	public void logInfo(Object obj, String msg) {
		log(obj, msg, LogKindEnum.INFO);
	}

	public void logWarning(Object obj, String msg) {
		log(obj, msg, LogKindEnum.WARNING);
	}

	public void logError(Object obj, String msg) {
		log(obj, msg, LogKindEnum.ERROR);
	}

	public void logTrace(Object obj, String msg) {
		log(obj, msg, LogKindEnum.TRACE);
	}

	/**
	 * The root function to print out the ECore model to the Rhapsody Log pane.
	 * @param modelElement
	 */
	public void printOut(EObject modelElement) {
		logInfo("--------------------");
		visited.clear();
		printOut(modelElement, null, 0);
		logInfo("--------------------");
	}

	/**
	 * Return indentation string for given indentation level consisting of 4*level blanks
	 * @param level
	 * @return
	 */
	private String indent(int level) {
		String indent = "    ";
		String str = "";
		while (level-- > 0) {
			str = indent + str;
		}
		return str;
	}

	/**
	 * This function prints out 'EObject' objects
	 * @param modelElement
	 * @param indentation
	 */
	private void printOut(EObject obj, EStructuralFeature eStructuralFeature, String indentation) {
		String objName = null;
		for (EAttribute attribute : obj.eClass().getEAllAttributes()) {
			if (attribute.getName().equals("name") && obj.eGet(attribute) instanceof String) {
				objName = (String)obj.eGet(attribute);
			}
		}
		if (eStructuralFeature != null) {
			String containment = (eStructuralFeature instanceof EReference && ((EReference)eStructuralFeature).isContainment() ? "(c)" : "");
			if (eStructuralFeature.isMany()) {
				logInfo(indentation + obj.eClass().getName() + ": '" + objName + "' [" + containment + eStructuralFeature.getName() + "]");
			} else {
				logInfo(indentation + containment + eStructuralFeature.getName() + "='" + obj.eClass().getName() + "::" + objName + "'");
			}
		} else {
			logInfo(indentation + obj.eClass().getName() + ": '" + objName + "'");
		}
	}

	/**
	 * This function prints out 'MModelElement' objects
	 * @param modelElement
	 * @param indentation
	 */
	private void printOut(MModelElement modelElement, EStructuralFeature eStructuralFeature, String indentation) {
		if (eStructuralFeature != null) {
			if (eStructuralFeature.isMany()) {
				logInfo(indentation + modelElement.eClass().getName() + ": '" + modelElement.getName() + "' [" + eStructuralFeature.getName() + "]");
			} else {
				logInfo(indentation + eStructuralFeature.getName() + "='" + modelElement.eClass().getName() + "::" + modelElement.getName() + "'");
			}
		} else {
			logInfo(indentation + modelElement.eClass().getName() + ": '" + modelElement.getName() + "'");
		}
	}

	/**
	 * This function prints out 'MModelElementRef' objects
	 * @param meRef
	 * @param indentation
	 */
	private void printOut(MModelElementRef meRef, String indentation) {
		logInfo(indentation + meRef.eClass().getName() + ": <'" + meRef.getStructuralFeatureName() + "', '" + meRef.getReference().eClass().getName() + "::" + meRef.getReference().getName() + "'>");
	}

	/**
	 * Recursively and reflectively navigate through the structural features and print out the model
	 * @param modelElement
	 * @param feature
	 * @param indentationLevel
	 */
	private void printOut(EObject modelElement, EStructuralFeature feature, int indentationLevel) {
		if (modelElement instanceof MModelElement) {
			printOut((MModelElement)modelElement, feature, indent(indentationLevel));
		} else {
			printOut(modelElement, feature, indent(indentationLevel));
		}
		if (!visited.contains(modelElement)) {
			// follow structural features only if this is the first visit to modelElement
			visited.add(modelElement);
			EList<EStructuralFeature> eStructuralFeatureList = modelElement.eClass().getEAllStructuralFeatures();
			for (EStructuralFeature eStructuralFeature : eStructuralFeatureList) {
				// We skip the structural features that are listed in the 'structuralFeaturesToSkip' list.
				if (eStructuralFeature.isMany()) {
					EList<Object> list = (EList<Object>) modelElement.eGet(eStructuralFeature);
					for (Object obj : list) {
						if (obj instanceof Character || obj instanceof String || obj instanceof Boolean
								|| obj instanceof Short || obj instanceof Integer || obj instanceof Long
								|| obj instanceof Float || obj instanceof Double) {
							// print out attribute
							logInfo(indent(indentationLevel + 1) + eStructuralFeature.getName() + "='" + obj.toString() + "'");
						} else if (obj instanceof MModelElementRef) {
								// print out model element reference
								printOut((MModelElementRef)obj, indent(indentationLevel + 1));
						} else if (obj instanceof EObject) {
							// recursively call the 'printOut(...)' function on obj
							printOut((EObject)obj, eStructuralFeature, indentationLevel + 1);
						}
					}
				} else {
					Object obj = modelElement.eGet(eStructuralFeature);
					if (obj instanceof Character || obj instanceof String || obj instanceof Boolean
						|| obj instanceof Short || obj instanceof Integer || obj instanceof Long
						|| obj instanceof Float || obj instanceof Double) {
					// print out attribute
					logInfo(indent(indentationLevel + 1) + eStructuralFeature.getName() + "='" + obj.toString() + "'");
					} else if (obj instanceof MModelElementRef) {
						// print out model element reference
						printOut((MModelElementRef)obj, indent(indentationLevel + 1));
					} else if (obj instanceof EObject) {
						// recursively call the 'printOut(...)' function on obj
						printOut((EObject)obj, eStructuralFeature, indentationLevel + 1);
					}
				}
			}
		}
	}
}
