
/****************************************************************************************
#  Copyright (c) 2011, ALES S.r.l. - Via C. Bartolomeo Piazza 13, 00161 Roma Italy      #
#  Developed with the sponsorship of the Defense Advanced Research Projects             #
#  Agency (DARPA).                                                                      #
#                                                                                       #
#  Delivered to the Government with Unlimited Rights under Contract #FA8650-10-C-7074.  #
#                                                                                       #
#  THE DATA IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR               #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,             #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE          #
#  AUTHORS, SPONSORS, DEVELOPERS, CONTRIBUTORS, OR COPYRIGHT HOLDERS BE LIABLE          #
#  FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF                   #
#  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE          #
#  DATA OR THE USE OR OTHER DEALINGS IN THE DATA.                                       #
****************************************************************************************/

package com.eu.ales.rpyplugins.utils;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.EStructuralFeature;

import com.eu.ales.rpyplugins.MClassifier;
import com.eu.ales.rpyplugins.MGeneralization;
import com.eu.ales.rpyplugins.MModelElement;
import com.eu.ales.rpyplugins.MModelElementRef;
import com.eu.ales.rpyplugins.MStereotype;
import com.eu.ales.rpyplugins.RpypluginsFactory;
import com.eu.ales.rpyplugins.RpypluginsPackage;

public class ModelUtils {
	public static Logger logger = null;
	public static final String GET = "get";
	public static final String PARSE = "parse";
	public static final String MODEL_ELEMENT_PREFIX = "M";
	public static final String UNSUPPORTED_MODEL_ELEMENTS = "UnsupportedModelElements";
	public static final String UNSUPPORTED_MODEL_ELEMENTS_RELATION = MODEL_ELEMENT_PREFIX.toLowerCase() + UNSUPPORTED_MODEL_ELEMENTS;
	public static final String MODEL_ELEMENT_REFS = "ModelElementRefs";
	public static final String MODEL_ELEMENT_REFS_RELATION = MODEL_ELEMENT_PREFIX.toLowerCase() + MODEL_ELEMENT_REFS;
	public static final String MODEL_ELEMENTS = "ModelElements";
	public static final String MODEL_ELEMENTS_RELATION = MODEL_ELEMENT_PREFIX.toLowerCase() + MODEL_ELEMENTS;
	public static final String RPYPLUGINS_MODEL_PACKAGE = "com.eu.ales.rpyplugins";

	public static final void setLogger(Logger logger) {
		ModelUtils.logger = logger;
	}

	/**
	 * Create model element of type className, if possible, otherwise throw exception.
	 * @param className
	 * @return
	 * @throws SecurityException
	 * @throws NoSuchMethodException
	 * @throws IllegalArgumentException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 */
	public static MModelElement createModelElementByReflection(String className) {
		try {
			String methodName = GET + className;
			Method method = RpypluginsPackage.eINSTANCE.getClass().getMethod(methodName);
			EClass modelElementEClass = (EClass) method.invoke(RpypluginsPackage.eINSTANCE);
			return (MModelElement) RpypluginsFactory.eINSTANCE.create(modelElementEClass);
		} catch (Exception e) {	
		} catch (Error e) {	
		}
		return null;
	}

	/**
	 * This function sets the relation 'relationName' between container and element
	 * This function assumes that:
	 * 1) if the initial character of 'relationName' is upper-cased, then the relation name is to be completed with a fixed predefined prefix
	 * 2) if the initial character of 'relationName' is lower-cased, then the relation name is to be left unchanged
	 * @param container
	 * @param element
	 * @param relationName
	 */
	public static void setECoreRelation(MModelElement container, MModelElement element, String relationName) {
		if (Character.isUpperCase(relationName.charAt(0))) {
			relationName = MODEL_ELEMENT_PREFIX.toLowerCase() + relationName;
		}
		EStructuralFeature eStructuralFeature = getContainmentOrReferenceRelation(container, relationName);
		setECoreRelation(container, element, eStructuralFeature);
	}

	/**
	 * This function adds or sets the 'MModelElement' element to the eStructuralFeature of container.
	 * Before setting it, the eStructuralFeature is cleaned (see comment of function 'cleanEStructuralFeature(...)').
	 * After setting the eStructuralFeature, an inverse reference 'MModelElementRef' object, containing a reference to container, 
	 * is created and added to the 'mModelElementRefs' relation of element
	 * @param container
	 * @param element
	 * @param eStructuralFeature
	 */
	public static void setECoreRelation(MModelElement container, MModelElement element, EStructuralFeature eStructuralFeature) {
		if (eStructuralFeature != null) {
			cleanEStructuralFeature(container, element, eStructuralFeature);
			if (eStructuralFeature.isMany()) {
				EList<EObject> list = (EList<EObject>) container.eGet(eStructuralFeature);
				list.add(element);
			} else {
				container.eSet(eStructuralFeature, element);
			}
			if (eStructuralFeature instanceof EReference && !((EReference)eStructuralFeature).isContainment()) {
				// add container to list of inverse references of element
				// applies only for non-containment relations
				MModelElementRef modelElementRef = RpypluginsFactory.eINSTANCE.createMModelElementRef();
				modelElementRef.setStructuralFeatureName(eStructuralFeature.getName());
				modelElementRef.setReference(container);
				element.getMModelElementRefs().add(modelElementRef);
			}
		} else if (logger != null) {
			logger.logTrace(container, "invalid null ECore structural feature");
		}
	}

	/**
	 * This function adds or sets element to the eStructuralFeature of container
	 * @param container
	 * @param element
	 * @param eStructuralFeature
	 */
	public static void setECoreAttribute(MModelElement container, Object element, EStructuralFeature eStructuralFeature) {
		if (element instanceof MModelElement) {
			setECoreRelation(container, (MModelElement)element, eStructuralFeature);
		} else {
			if (eStructuralFeature.isMany()) {
				EList<Object> list = (EList<Object>) container.eGet(eStructuralFeature);
				list.add(element);
			} else {
				container.eSet(eStructuralFeature, element);
			}
		}
	}

	/**
	 * Returns the EStructuralFeature with name 'relationName' of EObject container
	 * This is the ECore object corresponding to the 'relationName' 0..* reference relation of container
	 * @param container
	 * @param relationName
	 * @return
	 */
	public static  EStructuralFeature getContainmentOrReferenceRelation(EObject container, String relationName)
	{
		EStructuralFeature eStructuralFeature = container.eClass().getEStructuralFeature(relationName);
		if (eStructuralFeature == null) {
			List<EReference> eReferenceList = container.eClass().getEReferences();
			for (EReference eReference : eReferenceList) {
				if (eReference.equals(relationName)) {
					return eReference;
				}
			}
		}
		return eStructuralFeature;
	}

	/**
	 * This function does the following:
	 * 1) if the eStructuralFeature is a many relation containing element, element is removed from the relation
	 *    and the inverse reference to container is removed from the 'mModelElementRefs' relation of element
	 * 2) if the eStructuralFeature is not a many relation, the eStructuralFeature is reset to null 
	 *    and the inverse reference to container is removed from the 'mModelElementRefs' relation of 
	 *    the eStructuralFeature's referenced object
	 * @param container
	 * @param element
	 * @param eStructuralFeature
	 */
	public static void cleanEStructuralFeature(MModelElement container, MModelElement element, EStructuralFeature eStructuralFeature) {
		MModelElementRef modelElementRef = null;
		MModelElement me = null;
		if (eStructuralFeature.isMany()) {
			EList<EObject> list = (EList<EObject>) container.eGet(eStructuralFeature);
			if (list.contains(element)) {
				// remove element from relation
				list.remove(element);
				if (eStructuralFeature instanceof EReference && !((EReference)eStructuralFeature).isContainment()) {
					me = element;
				}
			}
		} else {
			if (eStructuralFeature instanceof EReference && !((EReference)eStructuralFeature).isContainment()) {
				// get the referenced object
				Object obj = container.eGet(eStructuralFeature);
				if (obj instanceof MModelElement) {
					me = (MModelElement)obj;
					// reset the relation to null
					container.eSet(eStructuralFeature, null);
				} else if (obj != null && logger != null) {
					logger.logTrace(container, "unexpected object of type '" + obj.getClass().getSimpleName() + "' as value of relation '" + eStructuralFeature.getName() + "'");
				}
			}
		}
		if (me != null) {
			// look for the inverse reference
			for (MModelElementRef mer : me.getMModelElementRefs()) {
				if (mer.getStructuralFeatureName().equals(eStructuralFeature.getName()) && mer.getReference() == container) {
					modelElementRef = mer;
					break;
				}
			}
			if (modelElementRef != null) {
				// remove the inverse reference
				me.getMModelElementRefs().remove(modelElementRef);
			} else if (logger != null) {
				logger.logTrace(container, "cannot find inverse reference of relation '" + eStructuralFeature.getName() + "' in element '" + me.getName() + "' of type '" + me.eClass().getName() + "'");
			}
		}
	}

	private static void assumption(boolean condition, String msg) {
		if (!condition) {
			throw new Error(msg);
		}
	}

	public static Set<MStereotype> getAllStereotypesAndTheirGeneralizations(MModelElement modelElement) {
		assumption(modelElement != null, "Invalid null model element argument");
		List<MStereotype> stereotypes = modelElement.getMStereotypes();
		Set<MStereotype> allStereotypes = new  HashSet<MStereotype>();
		allStereotypes.addAll(stereotypes);
		// now look for stereotype generalizations of the stereotypes of classifier
		List<MClassifier> classifiers = new ArrayList<MClassifier>();
		classifiers.addAll(stereotypes);
		Map<MClassifier,List<MClassifier>> allGeneralizations = getAllGeneralizations(classifiers);
		for (MStereotype stereotype : stereotypes) {
			List<MClassifier> allClassifiers = allGeneralizations.get(stereotype);
			if (allClassifiers != null) {
				for (MClassifier c : allClassifiers) {
					if (c instanceof MStereotype) {
						allStereotypes.add((MStereotype)c);
					}
				}
			} else if (logger != null) {
				logger.logTrace(modelElement, "no generalizations found for stereotypes of '" + modelElement.getName() + "'");
			}
		}
		return allStereotypes;
	}
	public static Map<MClassifier,List<MClassifier>> getAllGeneralizations(List<?extends MClassifier> classifiers) {
		assumption(classifiers != null, "Invalid null classifier argument");
		Map<MClassifier,List<MClassifier>> generalizations = new HashMap<MClassifier,List<MClassifier>>();
		for (MClassifier classifier : classifiers) {
			getAllGeneralizations(classifier, generalizations);
		}
		return generalizations;
	}
	public static Map<MClassifier,List<MClassifier>> getAllGeneralizations(MClassifier classifier) {
		assumption(classifier != null, "Invalid null classifier argument");
		Map<MClassifier,List<MClassifier>> generalizations = new HashMap<MClassifier,List<MClassifier>>();
		getAllGeneralizations(classifier, generalizations);
		return generalizations;
	}
	public static void getAllGeneralizations(MClassifier classifier, Map<MClassifier,List<MClassifier>> generalizations) {
		assumption(classifier != null, "Invalid null classifier argument");
		List<MClassifier> list = new ArrayList<MClassifier>();
		generalizations.put(classifier, list);
		List<MGeneralization> generalizationRelations = classifier.getMGeneralizations();
		for (MGeneralization generalizationRelation : generalizationRelations) {
			MClassifier baseClass = generalizationRelation.getBaseClass();
			getAllGeneralizations(baseClass, generalizations);
			list.add(baseClass);
			list.addAll(generalizations.get(baseClass));
		}
	}
	public static List<Class> getAllApplicableTo(MStereotype stereotype, Map<MClassifier,List<MClassifier>> generalizations) {
		if (generalizations == null || !generalizations.containsKey(stereotype)) {
			generalizations = getAllGeneralizations(stereotype);
		}
		List<Class> classes = new ArrayList<Class>();
		getApplicableTo(stereotype, classes);
		for (MClassifier classifier : generalizations.get(stereotype)) {
			if (classifier instanceof MStereotype) {
				getApplicableTo((MStereotype)classifier, classes);
			}
		}
		return classes;
	}
	public static void getApplicableTo(MStereotype stereotype, List<Class> classes) {
		String applicabilityString = stereotype.getOfMetaClass();
		if (applicabilityString != null && !applicabilityString.isEmpty()) {
			String[] names = applicabilityString.split(",");
			for (String name : names) {
				try {
					Class clas = Class.forName(RPYPLUGINS_MODEL_PACKAGE + "." + MODEL_ELEMENT_PREFIX + name);
					classes.add(clas);
				} catch (Exception e) {
				}
			}
		}
	}

	public static List<Class> getAllSuperClasses(Object obj) {
		return getAllSuperClasses(obj.getClass());		
	}

	public static List<Class> getAllSuperClasses(Class clas) {
		List<Class> allSuperClasses = new ArrayList<Class>();
		getAllSuperClasses(clas, allSuperClasses);
		return allSuperClasses;
	}

	public static void getAllSuperClasses(Class clas, List<Class> allSuperClasses) {
		Class directSuperClass = clas.getSuperclass();
		if (directSuperClass != null && !allSuperClasses.contains(directSuperClass)) {
			allSuperClasses.add(directSuperClass);
		}
		Class interfaces[] = clas.getInterfaces();
		if (interfaces != null) {
			for (Class c : interfaces) {
				if (!allSuperClasses.contains(c)) {
					allSuperClasses.add(c);
				}
			}
		}
		if (directSuperClass != null) {
			getAllSuperClasses(directSuperClass, allSuperClasses);
		}
		if (interfaces != null) {
			for (Class c : interfaces) {
				getAllSuperClasses(c, allSuperClasses);
			}
		}
	}
}
