
/****************************************************************************************
#  Copyright (c) 2011, ALES S.r.l. - Via C. Bartolomeo Piazza 13, 00161 Roma Italy      #
#  Developed with the sponsorship of the Defense Advanced Research Projects             #
#  Agency (DARPA).                                                                      #
#                                                                                       #
#  Delivered to the Government with Unlimited Rights under Contract #FA8650-10-C-7074.  #
#                                                                                       #
#  THE DATA IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR               #
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,             #
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE          #
#  AUTHORS, SPONSORS, DEVELOPERS, CONTRIBUTORS, OR COPYRIGHT HOLDERS BE LIABLE          #
#  FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF                   #
#  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE          #
#  DATA OR THE USE OR OTHER DEALINGS IN THE DATA.                                       #
****************************************************************************************/

package com.eu.ales.rpyplugins.utils;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

public class SystemCommands {
	public static final String BATCH_FILE_NAME = "commands";
	private static final Logger logger = Logger.getLogger(SystemCommands.class);

	private List<String> commands;

	public SystemCommands() {
		commands = new ArrayList<String>();
	}

	public void addCommand(String command) {
		if (command != null && !command.isEmpty()) {
			commands.add(command);
		} else {
			logger.logTrace("runCommand(String): cannot run null or empty command");
		}
	}

	public void addCommandsForFileConcatenationOnWindows(List<File> filesToConcatenate, File outputFile, boolean skipMissingFiles) {
		if (filesToConcatenate == null || filesToConcatenate.size() == 0) {
			throw new Error("no file to concatenate");
		}
		if (outputFile == null) {
			throw new Error("invalid output file");
		}
		String command = "copy /b /Y ";
		for (File file : filesToConcatenate) {
			if (file.exists()) {
				command += "\"" + file.getAbsolutePath() + "\"+";
			}
		}
		command = command.substring(0,command.length()-1);
		if (filesToConcatenate.contains(outputFile)) {
			String temporaryFilePath = "\"" + outputFile.getParentFile().getAbsolutePath() + File.separator + "temporary.tmp\"";
			addCommand(command + " " + temporaryFilePath);
			addCommand("move /Y " + temporaryFilePath + " \"" + outputFile + "\"");
		} else {
			addCommand(command + " " + outputFile);
		}
	}

	/**
	 * Commands are concatenated with the '&' batch command operator, that forces command sequentiality, 
	 * and are written to a script file which is returned
	 * @param commands
	 * @return
	 */
	public File createBatchFile() {
		if (!commands.isEmpty()) {
			FileWriter fstream = null;
			BufferedWriter writer = null;
			File commandFile = FileManager.getBatchFile(BATCH_FILE_NAME);
			try {
				fstream = new FileWriter(commandFile);
				writer = new BufferedWriter(fstream);
				writer.write(concatenateCommands());
				writer.flush();
				writer.close();
				return commandFile;
			} catch (IOException e) {
				logger.logError(e.getClass().getSimpleName() + " while opening file '" + commandFile.getAbsolutePath() + "' for writing: " + e.getMessage());
				return null;
			}
		} else {
			logger.logTrace("createBatchFile(String): empy commands, skipping script file creation");
		}
		return null;
	}

	public File createBatchFile(String batchFileRelativePath) {
		return FileManager.getBatchFile(batchFileRelativePath);
	}

	public int runBatchFile(String batchFileRelativePath) {
		return run(createBatchFile(batchFileRelativePath));
	}

	public int run(File scriptFile) {
		if (scriptFile != null) {
			Process process = run(scriptFile.getAbsolutePath());
			return process != null ? 0 : -1;
		} else {
			logger.logTrace("runScriptFile(File): cannot run null script file");
		}
		return -1;
	}

	public int run() {
		Process process = run(concatenateCommands());
		return process != null ? 0 : -1;
	}

	private Process run(String command) {
		if (command == null || command.isEmpty()) {
			throw new Error("Invalid operation: command string is null or empty");
		}
		Process process = null;
		try {
			logger.logTrace("Running system command '" + command + "'");
			process = Runtime.getRuntime().exec(command);
		} catch (Exception e) {
			logger.logError(e.getClass().getSimpleName() + " occurred while running system command '" + command + "': " + e.getMessage());
		}
		return process;
	}

	private String concatenateCommands() {
		String commandConcatenation = null;
		if (!commands.isEmpty()) {
			commandConcatenation = "";
			for (String command : commands) {
				commandConcatenation += command + " & ";
			}
			commandConcatenation = commandConcatenation.substring(0, commandConcatenation.length() - 2);
		} else {
			logger.logTrace("Empty commands, skipping operation");
		}
		return commandConcatenation;
	}
}
